__author__ = "Cameron Summers"
"""
InsuLearner: Estimating Insulin Pump Settings via Machine Learning

Code underlying this blog post:
https://www.cameronsummers.com/how_I_calculate_my_sons_insulin_pump_settings_with_machine_learning
"""

import sys
import os
import argparse
import datetime as dt

import warnings
import matplotlib.cbook
warnings.filterwarnings("ignore",category=matplotlib.MatplotlibDeprecationWarning)

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import scipy

from sklearn.linear_model import LinearRegression

from InsuLearner.tidepool.tidepool_user_model import TidepoolUser
from InsuLearner.tidepool.tidepool_api import TidepoolAPI
from InsuLearner.util import get_logger

from InsuLearner.carbohydrate_sensitivity_factor import estimate_csf

logger = get_logger(__name__)


def compute_aace_pump_settings(weight_kg, prepump_tdd):
    """
    Get pump settings using the American Association of Clinical Endocrinologists/American College of
    Endocrinology.

    Other references:
    https://diabetesed.net/wp-content/uploads/2019/09/Insulin-Pump-Calculations-Sept-2019-slides.pdf

    Review of insulin dosing formulas
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4960276/

    Args:
        weight_kg (float): weight in kg
        prepump_tdd (float): units of insulin per day, (CS: presuming this is a life-time average)

    Returns:
        (float, float, float): basal rate, carb insulin ratio, insulin sensitivity factor
    """
    tdd_method1 = weight_kg * 0.5
    tdd_method2 = prepump_tdd * 0.75
    starting_pump_tdd = (tdd_method1 + tdd_method2) / 2

    basal_rate = starting_pump_tdd * 0.5 / 24
    cir = 450.0 / starting_pump_tdd
    isf = 1700.0 / starting_pump_tdd

    return basal_rate, cir, isf


def pd_1d_series_to_X(series):
    """helper function"""
    return np.array(series.values.tolist())[:, np.newaxis]


def estimate_therapy_settings_from_window_stats_lr(aggregated_df,
                                                   K,
                                                   period_window_size_hours,
                                                   target_bg=110,
                                                   x="total_carbs",
                                                   y="total_insulin",
                                                   do_plots=True,
                                                   trained_model=None,
                                                   weight_scheme=None):
    """
    Fit the linear model and estimate settings
    """
    X_carbs = pd_1d_series_to_X(aggregated_df[x])
    y_insulin = aggregated_df[y]

    if weight_scheme is None:
        sample_weights = None
    else:
        if weight_scheme == "CGM Weighted":
            sample_weights = 1.0 / np.array(np.maximum(1.0, abs(target_bg - aggregated_df["cgm_geo_mean"])) / aggregated_df["cgm_percent_tir"]) * aggregated_df["cgm_percent_available"]
        elif weight_scheme == "Carb Uncertainty":
            sample_weights = scipy.stats.norm.pdf(aggregated_df["total_carbs"], aggregated_df["total_carbs"].mean(), aggregated_df["total_carbs"].std())
        else:
            raise Exception("Unknown weight scheme {}.".format(weight_scheme))

        nan_mask = np.isnan(sample_weights)
        X_carbs = X_carbs[~nan_mask]
        y_insulin = y_insulin[~nan_mask]
        sample_weights = sample_weights[~nan_mask]

        if do_plots:
            plt.figure()
            plt.title("Sample Weight Distribution")
            plt.hist(sample_weights)

    if trained_model is None:
        lm_carb_to_insulin = LinearRegression()
        lm_carb_to_insulin.fit(X_carbs, y_insulin, sample_weight=sample_weights)
    else:
        lm_carb_to_insulin = trained_model

    basal_insulin_estimate = lm_carb_to_insulin.intercept_

    r2_fit = lm_carb_to_insulin.score(X_carbs, y_insulin)
    logger.info(f"Linear Model: Fit R^2 {np.round(r2_fit, 2)}. Intercept {np.round(lm_carb_to_insulin.intercept_, 2)}. Slope g/U {np.round(1.0 / lm_carb_to_insulin.coef_, 2)}")

    cir_estimate_slope = 1 / lm_carb_to_insulin.coef_[0]
    isf_estimate_slope = cir_estimate_slope * K

    basal_rate_estimate_hourly = basal_insulin_estimate / period_window_size_hours

    logger.info("Total Period Basal={:.2f}U. (Mean %Daily Total: {:.2f}%)".format(basal_insulin_estimate, np.nanmean(basal_insulin_estimate / aggregated_df[y]) * 100))

    logger.info("\n\n\tSettings Estimates:\n")

    logger.info(f"\tEstimated CIR={round(cir_estimate_slope, 2)} g/U.")
    logger.info(f"\tEstimated Hourly Basal={round(basal_rate_estimate_hourly, 3)} U/hr")
    logger.info(f"\tCSF={round(K, 2)} mg/dL / g")
    logger.info(f"\tEstimated ISF={round(isf_estimate_slope, 2)} mg/dL/ U")

    settings = (cir_estimate_slope, isf_estimate_slope, basal_insulin_estimate, lm_carb_to_insulin, r2_fit, K)
    if do_plots:
        plot_aggregated_scatter(aggregated_df,
                                period_window_size_hours,
                                lr_model=lm_carb_to_insulin,
                                settings=settings,
                                plot_aace=True,
                                weight_scheme=weight_scheme)

    return settings


def plot_aggregated_scatter(aggregated_df, period_window_size_hours, lr_model=None, settings=None, plot_aace=True, weight_scheme=None):
    """
    Plot the linear model on the data.
    """
    period_window_size_days = period_window_size_hours / 24

    fig, ax = plt.subplots(figsize=(8, 8))
    win_start_dates_str = aggregated_df["start_date"].dt.strftime("%Y-%m-%d")
    win_end_dates_str = aggregated_df["end_date"].dt.strftime("%Y-%m-%d")
    plt.title(f"Insulin Prediction Modeling, Aggr Period={period_window_size_days} Days, {win_start_dates_str.values[0]} to {win_end_dates_str.values[-1]}, {len(aggregated_df)} data points")

    hue_col = "cgm_geo_mean"
    # hue_col = None

    vars_to_plot = ["total_carbs", "total_insulin", "cgm_geo_mean"]
    scatter_df = aggregated_df[vars_to_plot]

    sns.scatterplot(data=scatter_df, x="total_carbs", y="total_insulin", hue=hue_col, ax=ax)
    plt.figure()

    ax.set_ylim(0, aggregated_df["total_insulin"].max() * 1.1)
    ax.set_xlim(0, aggregated_df["total_carbs"].max() * 1.1)

    if settings:

        cir_estimate_slope, isf_estimate_slope, basal_insulin_estimate, lm_carb_to_insulin, r2_fit, K = settings
        basal_glucose_lr = -basal_insulin_estimate / lr_model.coef_[0]

        x1, y1 = basal_glucose_lr, 0
        x2, y2 = aggregated_df["total_carbs"].max(), lr_model.predict([[aggregated_df["total_carbs"].max()]])
        ax.plot([x1, x2], [y1, y2[0]], label="Insulin Prediction LR Model (Weights: {})".format(weight_scheme))

        ax.set_xlabel("Total Exogenous Glucose in Period T")
        ax.set_ylabel("Total Insulin in Period T")

        # Equations and Settings
        ax.text(0.6, 0.25, "y={:.4f}*x + {:.2f}, (R^2={:.2f})".format(lr_model.coef_[0], lr_model.intercept_, r2_fit), ha="left", va="top", transform=ax.transAxes)
        ax.text(0.6, 0.2, "CIR={:.2f} g/U \nBasal Rate={:.2f}U/hr \nISF={:.2f} mg/dL/U (K={:.2f})".format(cir_estimate_slope,
                                                                                                   basal_insulin_estimate / period_window_size_hours,
                                                                                                   isf_estimate_slope,
                                                                                                   K),
                                                                                                   ha="left", va="top", transform=ax.transAxes)

        # Stars
        ax.plot(0, basal_insulin_estimate, marker="*", markersize=12, color="green", label="Basal Insulin LR Estimate")
        ax.plot(basal_glucose_lr, 0, marker="*", markersize=12, color="orange", label="Endogenous Glucose LR Estimate")

        mean_insulin = aggregated_df["total_insulin"].mean()
        mean_carbs = aggregated_df["total_carbs"].mean()
        ax.plot(mean_carbs, mean_insulin, marker="*", markersize=12, color="red", label="Mean Insulin/Carbs")

        # Shades
        ax.fill_between([0, x2], [basal_insulin_estimate, basal_insulin_estimate], color="blue", alpha=0.2, label="Endogenous")
        ax.fill_between([0, x2], [basal_insulin_estimate, basal_insulin_estimate], [basal_insulin_estimate, y2[0]],
                        color="orange", alpha=0.2, label="Exogenous")

        # AACE line if using 1-day windows
        if plot_aace and period_window_size_hours == 1:
            tdd_mean = aggregated_df["total_insulin"].mean()
            aace_basal_insulin_estimate = tdd_mean / 2
            cir_aace = 450 / tdd_mean
            aace_basal_glucose_estimate = -aace_basal_insulin_estimate / (1/cir_aace)
            x2 = aggregated_df["total_carbs"].max()

            x1, y1 = (aace_basal_glucose_estimate, 0)
            y2 = 1.0 / (cir_aace) * x2 + aace_basal_insulin_estimate
            star_description = "AACE Basal Estimate (mean(TDD) / 2)"
            line_description = "AACE (*mean(TDD) only)"

            ax.plot([x1, x2], [y1, y2], label=line_description, color="gray", linestyle="--")
            ax.plot(0, aace_basal_insulin_estimate, marker="*", markersize=12, color="gray", label=star_description)

        ax.legend()

    plt.show()


def analyze_settings_lr(user, data_start_date, data_end_date,
                        K=12.5,
                        do_plots=True,
                        use_circadian_hour_estimate=True,
                        agg_period_window_size_hours=24,
                        agg_period_hop_size_hours=24,
                        weight_scheme=None):
    """
    Aggregate the data and fit a linear model
    """
    window_stats = user.compute_window_stats(data_start_date, data_end_date,
                                             use_circadian=use_circadian_hour_estimate,
                                             window_size_hours=agg_period_window_size_hours,
                                             hop_size_hours=agg_period_hop_size_hours,
                                             plot_hop_raw=False)
    window_df = pd.DataFrame(window_stats)

    logger.debug(f'Mean of CGM Mean, {np.round(window_df["cgm_mean"].mean(), 2)}')
    logger.debug(f'Mean of CGM Geo Mean, {np.round(window_df["cgm_geo_mean"].mean(), 2)}')
    logger.debug(f'Total Period Insulin Mean: {np.round(window_df["total_insulin"].mean(), 2)}')
    logger.debug(f"{len(window_df)} Data Rows")

    settings = estimate_therapy_settings_from_window_stats_lr(window_df,
                                                              K,
                                                              period_window_size_hours=agg_period_window_size_hours,
                                                              target_bg=110,
                                                              x="total_carbs",
                                                              y="total_insulin",
                                                              do_plots=do_plots,
                                                              weight_scheme=weight_scheme)

    cir_estimate, isf_estimate, basal_insulin_estimate, lr_model, lr_score, K = settings

    return cir_estimate, basal_insulin_estimate, isf_estimate, lr_score


def load_user_data(username, password, data_start_date, data_end_date, estimation_window_size_days):
    """
    Load the Tidepool user_obj data for the given user_obj and parameters.

    Args:
        username: tidepool username
        password: tidepool password
        data_start_date: start date of analysis
        data_end_date: end date of analysis
        estimation_window_size_days: size of estimation window in days

    Returns:
        TidepoolUser object
    """

    tp_api_obj = TidepoolAPI(username, password)
    user = TidepoolUser()
    user.load_from_api(tp_api_obj, data_start_date, data_end_date,
                       user_id=None,  # For a user_obj that is sharing their Tidepool account with this one
                       save_data=False)

    total_basal_days = user.get_num_days_span(data_type="basal")
    total_bolus_days = user.get_num_days_span(data_type="bolus")
    total_cgm_days = user.get_num_days_span(data_type="cgm")
    total_food_days = user.get_num_days_span(data_type="food")

    if [total_basal_days, total_bolus_days, total_cgm_days, total_food_days] != [estimation_window_size_days]:
        logger.warning(f"*** Warning *** : Num data days span not the size of estimation window size of {estimation_window_size_days} days")
        logger.warning(f"Basal Days Span {total_basal_days}. Bolus Days Span {total_bolus_days}. CGM Days Span {total_cgm_days}. Food Days Span {total_food_days}")

    user.analyze_duplicates(time_diff_thresh_sec=60 * 60)

    return user


def main():

    parser = argparse.ArgumentParser("InsuLearner: Estimate Insulin Pump Settings with Linear Regression")

    parser.add_argument("tp_username", type=str, help="Email username for Tidepool Account")
    parser.add_argument("tp_password", type=str, help="Password for Tidepool Account")
    parser.add_argument("-ht", "--height_inches", type=float, help="Your height in inches")
    parser.add_argument("-wt", "--weight_lbs", type=float, help="Your weight in pounds")
    parser.add_argument("-g", "--gender", choices=["male", "female"])
    parser.add_argument("--num_days", type=int, help="Number of days in the past to analyze data", default=60)
    parser.add_argument("--CSF", type=float, help="If entered, will use this CSF instead of estimating it from height and weight.")
    parser.add_argument("-eb", "--estimate_agg_boundaries", "-eb", action="store_true", default=True,
                        help="Use an autocorrelation-like algorithm to estimate aggregation boundaries to denoise the fit.")

    parser.add_argument("-aw", "--agg_period_window_size_hours", type=int,
                        help="The size in hours of each period to aggregate for fitting the model.", default=24 * 7)
    parser.add_argument("-ah", "--agg_period_hop_size_hours", "-ah", type=int,
                        help="The size in hours to hop each period for aggregation.", default=24)

    args = parser.parse_args()

    tp_username = args.tp_username
    tp_password = args.tp_password
    estimation_window_size_days = args.num_days
    height_inches = args.height_inches
    weight_lbs = args.weight_lbs
    gender = args.gender
    CSF = args.CSF
    estimate_agg_boundaries = args.estimate_agg_boundaries
    agg_period_window_size_hours = args.agg_period_window_size_hours
    agg_period_hop_size_hours = args.agg_period_hop_size_hours

    K = CSF
    if CSF is None:
        # Estimate CSF from blood volume based on height and weight
        K = estimate_csf(height_inches, weight_lbs, gender, metabolism_efficiency_percentage=0.23)
        logger.info(f"CSF estimated to be {np.round(K, 2)} for height_inches {height_inches}, weight_lbs {weight_lbs}, and gender {gender}")
    else:
        logger.info(f"Provided CSF={K}")

    # Get date info
    today = dt.datetime.now()
    data_start_date = today - dt.timedelta(days=estimation_window_size_days + 1)
    data_end_date = today - dt.timedelta(days=1)

    # Uncomment if specific dates desired
    # data_start_date = dt.datetime(year=2022, month=12, day=1)
    # data_end_date = dt.datetime(year=2023, month=1, day=26)

    logger.info(f"Running for dates {data_start_date} to {data_end_date}")

    # Load user_obj data into an object
    user_obj = load_user_data(tp_username, tp_password, data_start_date, data_end_date, estimation_window_size_days)

    # Run settings analysis
    analyze_settings_lr(user_obj,
                        data_start_date=data_start_date,
                        data_end_date=data_end_date,
                        K=K,
                        do_plots=True,
                        use_circadian_hour_estimate=estimate_agg_boundaries,
                        agg_period_window_size_hours=agg_period_window_size_hours,
                        agg_period_hop_size_hours=agg_period_hop_size_hours)


if __name__ == "__main__":
    main()


__author__ = "Cameron Summers"

"""
This estimates the Carbohydrate Sensitivity Factor (CSF) which is
the concentration level rise of glucose in the blood in mg/dL for
ingestion 1g carbohydrates. It is based on assumptions about
blood volume and carb metabolism.
"""


def estimate_blood_volume_d(height_inches, weight_lbs, gender):
    """
    Estimate blood volume based on height and weight based on Nadler's equation.

    Source: https://pubmed.ncbi.nlm.nih.gov/30252333/

    Args:
        height_inches: height in inches
        weight_lbs: weight in pounds

    Returns:
        blood volume in deciliters (float)
    """
    height_meters = height_inches / 39.37
    weight_kg = weight_lbs / 2.205

    if gender == "male":
        blood_volume_liters = (0.3669 * height_meters**3) + (0.03219 * weight_kg) + 0.6041
    elif gender == "female":
        blood_volume_liters = (0.3561 * height_meters**3) + (0.03308 * weight_kg) + 0.1833
    else:
        raise Exception(f"No equation for gender: {gender}.")

    blood_volume_dl = blood_volume_liters * 10

    return blood_volume_dl


def estimate_csf(height_inches, weight_lbs, gender, metabolism_efficiency_percentage=0.23):
    """
    Estimates the rise in blood glucose concentration based on dissolving
    1000g of glucose into blood.

    23% is a hypothesis:

    In the equations for CIR and ISF from American Association of Endocrinology,
     their CIR and ISF - which is a population average - places CSF average at 3.8 mg/dL / g.
     The blood volume using Nadler equation of an average height and weight person in U.S. is about 55dL.
     100% efficiency would be 1000mg / 55dL = 18.2 mg/dL. So efficiency is 3.8 / 18.2 ~ 21%.

     Corroborating: My 6yo son's blood volume is estimated at about 19.1 dL.
     I estimated his CSF at about 12.5 mg/dL / g.
     100% efficiency for him would be 1000 / 19.1 dL = 52.36 mg/dL.
     So his efficiency is 12.5 / 52.36 = 23.9%.

     Taking a conservative approach (leads to less strong ISF), the default is set to 23%.

     Height and Weights: https://www.cdc.gov/nchs/data/nhsr/nhsr122-508.pdf

    Args:
        height_inches: height in inches
        weight_lbs: weight in pounds
        gender: gender
        metabolism_efficiency_percentage: percentage of carbs metabolized into blood

    Returns:
        Carbohydrate Sensitivity Factor (float)
    """
    blood_volume_dl = estimate_blood_volume_d(height_inches, weight_lbs, gender)
    carbs_mg = 1000
    CSF = carbs_mg / blood_volume_dl * metabolism_efficiency_percentage

    return CSF



__author__ = "Cameron Summers"

# -*- coding: utf-8 -*-
"""
Utilities for downloading projects from Tidepool API

Reference: https://developer.tidepool.org/tidepool-api/index/
"""

import os
import datetime as dt
import sys
import requests

import logging
from InsuLearner.util import DATESTAMP_FORMAT,  get_logger

logger = get_logger(__name__)


class TidepoolAPI(object):
    """
    Class representing a user_obj with a Tidepool account.

    # TODO: Add checks and enforcement for order of events
    # TODO: Add helper functions for getting earlier/latest data
    """

    def __init__(self, username, password):

        self.login_url = "https://api.tidepool.org/auth/login"

        self.user_data_url = "https://api.tidepool.org/data/{user_id}"
        self.logout_url = "https://api.tidepool.org/auth/logout"
        self.users_sharing_to_url = "https://api.tidepool.org/metadata/users/{user_id}/users"
        self.users_sharing_with_url = "https://api.tidepool.org/access/groups/{user_id}"
        self.invitations_url = "https://api.tidepool.org/confirm/invitations/{user_id}"
        self.accept_invitations_url = "https://api.tidepool.org/confirm/accept/invite/{observer_id}/{user_id}"
        self.user_notes_url = "https://api.tidepool.org/message/notes/{user_id}"

        self.username = username
        self.password = password

        self._login_user_id = None
        self._login_headers = None

    def _check_login(func):
        """
        Decorator for enforcing login.
        """
        def is_logged_in(self, *args, **kwargs):
            if self._login_headers is None or self._login_user_id is None:
                raise Exception("Not logged in.")
            return func(self, *args, **kwargs)
        return is_logged_in

    def _check_http_error(func):
        """
        Decorator to batch handle failed http requests.
        """
        def response_is_ok(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except requests.HTTPError as e:
                logger.info("Failed request. HTTPError: {}".format(e))
        return response_is_ok

    def login(self):
        """
        Login to Tidepool API
        """
        login_response = requests.post(self.login_url, auth=(self.username, self.password))

        xtoken = login_response.headers["x-tidepool-session-token"]
        user_id_master = login_response.json()["userid"]

        self._login_user_id = user_id_master
        self._login_headers = {
            "x-tidepool-session-token": xtoken,
            "Content-Type": "application/json"
        }

    @_check_http_error
    @_check_login
    def logout(self):
        """
        Logout of Tidepool API

        Args:
            auth:

        Returns:

        """
        logout_response = requests.post(self.logout_url, auth=(self.username, self.password))
        logout_response.raise_for_status()

    @_check_login
    def get_pending_observer_invitations(self):
        """
        Get pending invitations that have been sent to an observer.

        Args:
            user_id_observer:
            headers:

        Returns:
            list of invitation json objects
        """
        try:
            invitations_url = self.invitations_url.format(**{"user_id": self._login_user_id})
            invitations_response = requests.get(invitations_url, headers=self._login_headers)
            invitations_response.raise_for_status()

            pending_invitations_json = invitations_response.json()
        except requests.HTTPError:
            pending_invitations_json = []

        return pending_invitations_json

    @_check_login
    def accept_observer_invitations(self):
        """
        Get pending invitations sent to an observer and accept them.

        Args:
            user_id_observer:
            headers:

        Returns:
            (list, list)
            pending
        """
        pending_invitations_json = self.get_pending_observer_invitations()

        total_invitations = len(pending_invitations_json)
        logger.info("Num pending invitations {}".format(total_invitations))

        invitation_accept_failed = []

        for i, invitation in enumerate(pending_invitations_json):

            try:
                share_key = invitation["key"]
                user_id = invitation["creatorId"]
                accept_url = self.accept_invitations_url.format(**{"observer_id": self._login_user_id, "user_id": user_id})

                accept_response = requests.put(accept_url, headers=self._login_headers, json={"key": share_key})
                accept_response.raise_for_status()

            except requests.HTTPError as e:
                invitation_accept_failed.append((e, invitation))

            if i % 20 == 0:
                num_failed = len(invitation_accept_failed)
                logger.info("Accepted {}. Failed {}. Out of {}".format(i - num_failed, num_failed, total_invitations))

        return pending_invitations_json, invitation_accept_failed

    @_check_http_error
    @_check_login
    def get_user_event_data(self, start_date, end_date, observed_user_id=None):
        """
        Get health event data for user_obj. TODO: Make more flexible

        Args:
            start_date (dt.datetime): Start date of data, inclusive
            end_date (dt.datetime): End date of data, inclusive of entire day
            observed_user_id (str): Optional id of observed user_obj if login id is clinician/study

        Returns:
            list: List of events as objects
        """
        user_id = self._login_user_id
        if observed_user_id:
            user_id = observed_user_id

        start_date_str, end_date_str = self.get_date_filter_string(start_date, end_date)

        user_data_base_url = self.user_data_url.format(**{"user_id": user_id})
        user_data_url = "{url_base}?startDate={start_date}&endDate={end_date}&dexcom=true&medtronic=true&carelink=true".format(**{
            "url_base": user_data_base_url,
            "end_date": end_date_str,
            "start_date": start_date_str,
        })

        data_response = requests.get(user_data_url, headers=self._login_headers)
        data_response.raise_for_status()
        user_event_data = data_response.json()

        return user_event_data

    @_check_http_error
    @_check_login
    def get_users_sharing_to(self):
        """
        Get a list of users the login id is sharing data to. The login id is typically
        a patient and the user_obj list is clinicians or studies.

        Returns:
            list: List of users as objects
        """

        user_metadata_url = self.users_sharing_to_url.format(**{
            "user_id": self._login_user_id
        })

        metadata_response = requests.get(user_metadata_url, headers=self._login_headers)
        metadata_response.raise_for_status()
        users_sharing_to = metadata_response.json()

        return users_sharing_to

    @_check_http_error
    @_check_login
    def get_users_sharing_with(self):
        """
        Get a list of users the login id is observing. The login id is typically the
        clinician or study and the user_obj list is patients.

        Returns:
            dict: List of users as objects
        """

        users_sharing_with_url = self.users_sharing_with_url.format(**{
            "user_id": self._login_user_id
        })
        users_sharing_with_response = requests.get(users_sharing_with_url, headers=self._login_headers)
        users_sharing_with_response.raise_for_status()
        users_sharing_with_json = users_sharing_with_response.json()

        return users_sharing_with_json

    @_check_http_error
    @_check_login
    def get_user_metadata(self, observed_user_id):

        user_metadata_url = self.users_sharing_to_url.format(**{
            "user_id": self._login_user_id
        })
        metadata_response = requests.get(user_metadata_url, headers=self._login_headers)

        metadata_response.raise_for_status()
        user_metadata = metadata_response.json()

        return user_metadata

    @_check_http_error
    @_check_login
    def get_notes(self, start_date, end_date, observed_user_id=None):
        """
        Get notes for a user_obj.
        """
        user_id = self._login_user_id
        if observed_user_id:
            user_id = observed_user_id

        start_date_str, end_date_str = self.get_date_filter_string(start_date, end_date)

        base_notes_url = self.user_notes_url.format(**{"user_id": user_id})
        notes_url = "{url_base}?startDate={start_date}&endDate={end_date}".format(
            **{
                "url_base": base_notes_url,
                "end_date": end_date_str,
                "start_date": start_date_str,
            })
        notes_response = requests.get(notes_url, headers=self._login_headers)

        try:
            notes_response.raise_for_status()
            notes_data = notes_response.json()
        except requests.HTTPError:
            notes_data = []

        return notes_data

    def get_date_filter_string(self, start_date, end_date):
        """
        Get string representations for date filters.

        Args:
            start_date dt.DateTime: start date
            end_date dt.Datetime: end date

        Returns:
            (str, str): start and end date strings
        """

        start_date_str = start_date.strftime(DATESTAMP_FORMAT) + "T00:00:00.000Z"
        end_date_str = end_date.strftime(DATESTAMP_FORMAT) + "T23:59:59.999Z"

        return start_date_str, end_date_str

    @_check_login
    def get_login_user_id(self):
        """
        Get the id of the user_obj logged in.
        Returns:
            str: user_obj id
        """
        return self._login_user_id







import os
from collections import OrderedDict, defaultdict
import datetime as dt
import json
from operator import itemgetter

import numpy as np
from scipy.stats import gmean, gstd

import logging

from InsuLearner.util import (
    API_DATA_TIMESTAMP_FORMAT, parse_tidepool_api_date_str, create_user_dir, DATESTAMP_FORMAT, get_recursively, get_logger
)
from InsuLearner.visualize_user_data import (
    plot_raw_data, plot_daily_stats
)

logger = get_logger(__name__)


# ===============  Tidepool API Objects ================#

class TidepoolMeasurement(object):
    """
    Object to represent a measurement of something in the Tidepool system.
    """

    def __init__(self, value, units, mid, upload_id, source):
        if not isinstance(value, float):
            try:
                value = float(value)
            except Exception as e:
                logger.debug("Exception: TidepoolMeasurement")
                raise e

        self.value = value
        self.units = units
        self.id = mid
        self.upload_id = upload_id
        self.source = source

    def get_value(self):
        return self.value

    def get_units(self):
        return self.units

    def is_equal(self, other, value_diff_threshold):
        return self.__class__ == other.__class__ and \
               abs(self.value - other.value) < value_diff_threshold and \
               self.units == other.units


class TidepoolGlucoseMeasurement(TidepoolMeasurement):

    def __init__(self, value, units, mid, upload_id, source):
        super().__init__(value, units, mid, upload_id, source)

        if units == "mmol/L":
            self.value *= 18.0182
            self.units = "mg/dL"


class TidepoolManualGlucoseMeasurement(TidepoolGlucoseMeasurement):

    def __init__(self, value, units, mid, upload_id, source):

        super().__init__(value, units, mid, upload_id, source)


class TidepoolCGMGlucoseMeasurement(TidepoolGlucoseMeasurement):

    def __init__(self, value, units, mid, upload_id, source):

        super().__init__(value, units, mid, upload_id, source)


class TidepoolFood(TidepoolMeasurement):

    def __init__(self, value, units, mid, upload_id, source, duration_sec=None):
        super().__init__(value, units, mid, upload_id, source)
        self.duration_sec = duration_sec


class TidepoolBasal(TidepoolMeasurement):

    def __init__(self, value, units, duration_hours, delivery_type, mid, upload_id, source):
        super().__init__(value, units, mid, upload_id, source)

        self.duration_hours = duration_hours
        self.delivery_type = delivery_type

    def get_duration_hours(self):
        return self.duration_hours

    def is_equal(self, other, value_diff_threshold):

        super().is_equal(other, value_diff_threshold) and \
        self.duration_hours == other.duration_hours and \
        self.delivery_type == other.delivery_type


class TidepoolBolus(TidepoolMeasurement):

    def __init__(self, value, units, mid, upload_id, source):
        super().__init__(value, units, mid, upload_id, source)


class TidepoolSquareBolus(TidepoolBolus):

    def __init__(self, value, units, duration, mid, upload_id, source):
        super().__init__(value, units, mid, upload_id, source)

        self.duration = duration


class TidepoolTimeChange():

    def __init__(self, timechange_event, mid, upload_id, source):

        self.timechange_event = timechange_event


class TidepoolReservoirChange():

    def __init__(self, event_data, mid, upload_id, source):

        self.data = event_data


class TidepoolNote():
    """
    A Tidepool Note
    """

    def __init__(self):

        self.note_datetime = None
        self.created_datetime = None
        self.message = None

        self.sensor_change_algo_map = {
            "v1": self._is_sensor_change_v1
        }

    def from_raw(self, note):
        """
        Constructor from raw json.

        Args:
            note (dict): raw json from Tidepool API

        Returns:
            self
        """

        self.note_datetime = parse_tidepool_api_date_str(note["timestamp"])
        self.created_datetime = parse_tidepool_api_date_str(note["createdtime"])
        self.message = note["messagetext"]

        return self

    def is_type(self, note_type, algo_version="v1"):

        return self.sensor_change_algo_map[algo_version](note_type)

    def _is_sensor_change_v1(self, note_type):

        return note_type in self.message

    def has_tag(self, tag_string, case_match=False):

        if case_match:
            return tag_string in self.message
        else:
            return tag_string.lower() in self.message.lower()

    def get_tags(self):
        raise NotImplementedError


class TidepoolWizard():

    def __init__(self, wizard_event, mid, upload_id, source):

        self.raw_data = wizard_event
        self.id = mid
        self.upload_id = upload_id


# ========== Tidepool User ================

class TidepoolUser(object):
    """
    Class representing a Tidepool user_obj from their data.
    """
    def __init__(self, device_data_json=None, notes_json=None, api_version="v1"):
        """
        Args:
            device_data_json (list): list of event data of any kind in Tidepool API
            api_version (str): parser version to user_obj
        """
        self._notes_filename = "notes.json"
        self._event_data_filename = "event_data.json"
        self._creation_meta_filename = "creation_metadata.json"

        self.unparsed_events = defaultdict(int)

        self.device_data_json = device_data_json
        self.notes_json = notes_json
        self.api_version = api_version

        self.data_parser_map = {
            "v1": self.parse_data_json_v1
        }

        self.notes_parser_map = {
            "v1": self.parse_notes_json_v1
        }

        # Manage device data in separate timelines
        self.basal_timeline = OrderedDict()
        self.bolus_timeline = OrderedDict()
        self.food_timeline = OrderedDict()
        self.cgm_timeline = OrderedDict()
        self.wizard_timeline = OrderedDict()
        self.time_change_timeline = OrderedDict()
        self.pump_settings_timeline = OrderedDict()
        self.reservoir_change_timeline = OrderedDict()

        # Notes data
        self.note_timeline = OrderedDict()

        self.num_absorption_times = 0 # tmp for carb duration counts

        self._parse()

    def _parse(self):

        if self.device_data_json is not None and len(self.device_data_json) > 0:
            self.data_parser_map[self.api_version]()

        try:
            if self.notes_json is not None and len(self.notes_json) > 0:
                self.notes_parser_map[self.api_version]()
        except Exception as e:
            logger.warning("Failed to parse notes: {}".format(e))

    def load_from_api(self, tp_api_obj, start_date, end_date, user_id=None, save_data=True, user_group_name=""):
        """
        Use Tidepool API to download Tidepool user_obj data

        Args:
            tp_api_obj (TidepoolAPI): credentialed api object
            save_dir (str): directory where the user_obj data will be stored
            start_date (dt.DateTime): start date of data collection
            end_date dt.DateTime: end date of data collection
            user_id (str): Optional user_obj id if the login credentials are an observer
        """
        tp_api_obj.login()

        # Create directory based on user_obj id whose data this is
        user_id_of_data = user_id
        if user_id_of_data is None:
            user_id_of_data = tp_api_obj.get_login_user_id()

        if save_data:
            save_dir = create_user_dir(user_id_of_data, start_date, end_date, user_group_name=user_group_name)

        # Download and save events
        user_event_json = tp_api_obj.get_user_event_data(start_date, end_date, observed_user_id=user_id)
        if save_data:
            json.dump(user_event_json, open(os.path.join(save_dir, self._event_data_filename), "w"))

        # Download and save notes
        notes_json = tp_api_obj.get_notes(start_date, end_date, observed_user_id=user_id)
        if save_data:
            json.dump(notes_json, open(os.path.join(save_dir, self._notes_filename), "w"))

        # TODO: add profile metadata

        tp_api_obj.logout()

        # Document this operation and save
        creation_metadata = {
            "date_created": dt.datetime.now().isoformat(),
            "api_version": "v1",
            "data_start_date": start_date.strftime(DATESTAMP_FORMAT),
            "data_end_date": end_date.strftime(DATESTAMP_FORMAT),
            "user_id": user_id,
            "observer_id": tp_api_obj.get_login_user_id()
        }
        if save_data:
            json.dump(creation_metadata, open(os.path.join(save_dir, self._creation_meta_filename), "w"))

        self.api_version = "v1"
        self.device_data_json = user_event_json
        self.notes_json = notes_json

        self._parse()

    def load_from_dir(self, path_to_user_data_dir):
        """
        Load stored data and parse.
        """
        # TODO: require device data?
        event_data_json = json.load(open(os.path.join(path_to_user_data_dir, self._event_data_filename)))
        self.device_data_json = event_data_json

        try:
            notes_json = json.load(open(os.path.join(path_to_user_data_dir, self._notes_filename)))
            self.notes_json = notes_json
        except FileNotFoundError:
            logger.warning("No Notes data found.")
            self.notes_json = None

        try:
            creation_meta_json = json.load(open(os.path.join(path_to_user_data_dir, self._creation_meta_filename)))
            creation_meta_json["data_start_date"] = parse_tidepool_api_date_str(creation_meta_json["data_start_date"])
            creation_meta_json["data_end_date"] = parse_tidepool_api_date_str(creation_meta_json["data_end_date"])
            self.id = creation_meta_json["user_id"]
        except FileNotFoundError:
            logger.warning("No creation metadata found.")

            self.id = None

        self._parse()

    def add_event_to_timeline(self, timeline, event_datetime, event):

        # Enforce order of datetime keys
        if len(timeline) > 0:
            last_datetime_added = next(iter(timeline))
            if event_datetime > last_datetime_added:
                raise Exception("Times for events not in order")

        timeline[event_datetime] = event

    def parse_data_json_v1(self):
        """
        Parse the json list into different event types

        Assumes and asserts device data is already sorted, but checks for direction of sort.
        """
        # time example: "2020-01-02T23:15:12.611Z"

        first_event_datetime = parse_tidepool_api_date_str(self.device_data_json[0]["time"])
        last_event_datetime = parse_tidepool_api_date_str(self.device_data_json[-1]["time"])

        if first_event_datetime < last_event_datetime:
            self.device_data_json.reverse()

        for i, event in enumerate(self.device_data_json):

            try:

                event_type = event.get("type")

                mid = event.get("id")
                upload_id = event.get("uploadId")
                origin = event.get("origin")
                if origin is not None:
                    source_name = origin.get("name", "")
                else:
                    source_name = event.get("deviceId", "")

                time_str = event["time"]
                event_datetime = parse_tidepool_api_date_str(time_str)

                if event_type == "smbg":
                    smbg = TidepoolManualGlucoseMeasurement(event["value"], event["units"], mid, upload_id, source_name)

                elif event_type == "cbg":

                    cbg = TidepoolCGMGlucoseMeasurement(event["value"], event["units"], mid, upload_id, source_name)
                    self.add_event_to_timeline(self.cgm_timeline, event_datetime, cbg)

                elif event_type == "food":

                    carb_key = "carbohydrate"
                    if carb_key not in event["nutrition"]:
                        carb_key = "carbohydrates"

                    value = float(event["nutrition"][carb_key]["net"])
                    units = event["nutrition"][carb_key]["units"]

                    duration_sec = None
                    if event.get("payload") and event["payload"].get("com.loudnate.CarbKit.HKMetadataKey.AbsorptionTimeMinutes"):
                        duration_sec = int(event["payload"].get("com.loudnate.CarbKit.HKMetadataKey.AbsorptionTimeMinutes"))
                        self.num_absorption_times += 1

                    food = TidepoolFood(value, units, mid, upload_id, source_name, duration_sec=duration_sec)
                    self.add_event_to_timeline(self.food_timeline, event_datetime, food)

                elif event_type == "basal":

                    try:
                        duration_ms = float(event["duration"])
                    except Exception as e:
                        logger.debug("Can't convert basal duration to float.")
                        raise e

                    duration_hours = duration_ms / 1000.0 / 3600

                    if event.get("rate") in [None, ""] and event["deliveryType"] == "suspend":
                        basal = TidepoolBasal(0, "U/hr", duration_hours, "suspend", mid, upload_id, source_name)
                    else:
                        basal = TidepoolBasal(event["rate"], "U/hr", duration_hours, event["deliveryType"], mid, upload_id, source_name)

                    self.add_event_to_timeline(self.basal_timeline, event_datetime, basal)

                elif event_type == "bolus":

                    sub_type = event["subType"]

                    if sub_type == "square":
                        bolus = TidepoolSquareBolus(event["extended"], "Units", event["duration"], mid, upload_id, source_name)
                    else:
                        bolus = TidepoolBolus(event["normal"], "Units", mid, upload_id, source_name)

                    self.add_event_to_timeline(self.bolus_timeline, event_datetime, bolus)

                    # On pumps where carbs are entered and stored with bolus
                    # TODO: make sure this doesn't clash with food events
                    if event.get("carbInput") not in ["", None, "0.0"]:
                        logger.debug("Found 'stitched' bolus/wizard. Adding carb.")
                        carb_input = float(event["carbInput"])
                        food = TidepoolFood(carb_input, "g", mid, upload_id, source_name)
                        self.add_event_to_timeline(self.food_timeline, event_datetime, food)

                elif event_type == "insulin":

                    try:
                        dose_total = event["dose.total"]
                    except KeyError:
                        dose_total = event["dose"]["total"]

                    bolus = TidepoolBolus(dose_total, "Units", mid, upload_id, source_name)  # manual
                    self.bolus_timeline[event_datetime] = bolus

                    self.add_event_to_timeline(self.bolus_timeline, event_datetime, bolus)

                elif event_type == "wizard":

                    wizard_event = TidepoolWizard(event, mid, upload_id, source_name)
                    self.add_event_to_timeline(self.wizard_timeline, event_datetime, wizard_event)

                    if wizard_event.raw_data.get("carbInput") is not None:
                        logger.debug("Wizard Carb Event!")
                        food = TidepoolFood(wizard_event.raw_data["carbInput"], "g", mid, upload_id, source_name)
                        self.add_event_to_timeline(self.food_timeline, event_datetime, food)

                elif event_type == "deviceEvent":

                    sub_type = event.get("subType")

                    if sub_type == "status":
                        # Suspended pump
                        self.unparsed_events["{}-{}".format(event_type, sub_type)] += 1
                    elif sub_type == "prime":
                        # Prime pump
                        self.unparsed_events["{}-{}".format(event_type, sub_type)] += 1
                    elif sub_type == "timeChange":
                        time_change = TidepoolTimeChange(event, mid, upload_id, source_name)
                        self.add_event_to_timeline(self.time_change_timeline, event_datetime, time_change)
                    elif sub_type == "reservoirChange":
                        reservoir_change = TidepoolReservoirChange(event, mid, upload_id, source_name)
                        self.add_event_to_timeline(self.reservoir_change_timeline, event_datetime, reservoir_change)
                    else:
                        self.unparsed_events["{}-{}".format(event_type, sub_type)] += 1
                        logger.debug("Device event not parsed. Sub type: {}".format(sub_type))

                elif event_type == "pumpSettings":
                    self.add_event_to_timeline(self.pump_settings_timeline, event_datetime, event)

                elif event_type == "upload":
                    self.unparsed_events["{}".format(event_type)] += 1
                elif event_type == "cgmSettings":
                    self.unparsed_events["{}".format(event_type)] += 1
                elif event_type == "physicalActivity":
                    self.unparsed_events["{}".format(event_type)] += 1
                else:
                    raise Exception("Unknown event type")

            except Exception as e:
                logger.warning("Dropped an event of type {} due to error: {}".format(event_type, e))

        if len(self.unparsed_events) > 0:
            logger.warning("Unparsed events: {}".format(self.unparsed_events))

    def parse_notes_json_v1(self):
        """
        Parse the Tidepool notes json.
        """
        for note in self.notes_json.get("messages", []):

            note_obj = TidepoolNote().from_raw(note)
            self.note_timeline[note_obj.note_datetime] = note_obj

    def get_num_days_span(self, data_type="basal"):
        """
        Count the number of days between the beginning and end of an event timeline.
        """
        if data_type == "cgm":
            timeline = self.cgm_timeline
        elif data_type == "bolus":
            timeline = self.bolus_timeline
        elif data_type == "basal":
            timeline = self.basal_timeline
        elif data_type == "food":
            timeline = self.food_timeline
        else:
            raise Exception(f"Data type unknown: {data_type}")

        data_end_date = list(timeline.items())[0][0]
        data_start_date = list(timeline.items())[-1][0]

        total_days_in_data = int((data_end_date - data_start_date).total_seconds() / 3600 / 24)

        return total_days_in_data

    def analyze_duplicates(self, time_diff_thresh_sec=5):

        self.cgm_timeline, found_dupe_cgm = self.deduplicate_timeline(self.cgm_timeline,
                                                                          value_diff_min_thres=1.0,
                                                                          time_diff_thresh_sec=60 * 60,  # within one hour and of different source
                                                                          description="CBG")

        self.bolus_timeline, found_dupe_bolus = self.deduplicate_timeline(self.bolus_timeline,
                                    value_diff_min_thres=1e-6,
                                  time_diff_thresh_sec=time_diff_thresh_sec,
                                  description="Bolus")

        self.basal_timeline, found_dupe_basal = self.deduplicate_timeline(self.basal_timeline,
                                  value_diff_min_thres=1e-6,
                                  time_diff_thresh_sec=time_diff_thresh_sec,
                                  description="Basal")

        self.food_timeline, found_dupe_food = self.deduplicate_timeline(self.food_timeline,
                                  value_diff_min_thres=1e-6,
                                  time_diff_thresh_sec=time_diff_thresh_sec,
                                  description="Food")

    def deduplicate_timeline(self, timeline, value_diff_min_thres, time_diff_thresh_sec=60, description=None):
        """
        Assumes the events are in order and that duplicates are next to one another.
        """
        logger.warning("De-duplicating {} data...".format(description))

        duplicate_candidates = []
        deduped_timeline = OrderedDict()
        found_dupes = False

        for i, (candidate_dt, candidate_event) in enumerate(list(timeline.items())):

            if i == 0:
                base_event = candidate_event
                base_dt = candidate_dt
                deduped_timeline[candidate_dt] = candidate_event
                continue

            if self.is_duplicate_event_basic(candidate_event, candidate_dt, base_event, base_dt, value_diff_min_thres, time_diff_thresh_sec):
                duplicate_candidates.append((base_dt, base_event, candidate_dt, candidate_event))
            else:
                deduped_timeline[candidate_dt] = candidate_event

                base_event = candidate_event
                base_dt = candidate_dt

        if len(duplicate_candidates) > 0:
            found_dupes = True
            logger.info(f"In {description} Data: Total events {len(timeline)}. Num dupes {len(duplicate_candidates)}")
            # deduped_timeline = timeline  # TODO expose optionally don't dedupe after analysis
            logger.info(f"Num {description} events after dedupe: {len(deduped_timeline)}")
        else:
            logger.info(f"Zero duplicates found in {description} data.")

        return deduped_timeline, found_dupes

    def is_duplicate_event_basic(self, event1, event1_dt, event2, event2_dt, value_diff_thresh, time_diff_thresh_different_sources):
        """
        According to backend TP engineers, duplicates are likely due to similar events from multiple sources.

        So if an event is from a different source, is close in specified value, and close in specified time, it is
        considered a duplicate.
        """

        if TidepoolMeasurement in type(event1).__mro__:
            seconds_between = abs(event1_dt - event2_dt).total_seconds()

            is_same_source = event1.source == event2.source
            is_same_event = event1.is_equal(event2, value_diff_thresh)

            if is_same_source and is_same_event:
                is_same_time = seconds_between <= 5  # within 5 seconds on the same device
            elif not is_same_source and is_same_event:
                is_same_time = seconds_between <= time_diff_thresh_different_sources
            else:
                is_same_time = False

            is_dupe = is_same_event and is_same_time and is_same_source

        elif isinstance(event1, TidepoolWizard):
            pass
        else:
            raise Exception()

        return is_dupe

    def get_insulin_stats(self, start_date, end_date):
        """
        Get the sum of insulin with the two datetimes, inclusive.

        Args:
            start_date (dt.DateTime): start date
            end_date (dt.DateTime): end date

        Returns:
            (float, int, float, int): sum and counts of bolus and basal
        """

        insulin_stats = {
            "total_bolus": 0.0,
            "num_bolus_events": 0,
            "total_basal": 0.0,
            "num_basal_events": 0
        }

        for time, bolus in self.bolus_timeline.items():
            if start_date <= time <= end_date:
                insulin_stats["total_bolus"] += bolus.get_value()
                insulin_stats["num_bolus_events"] += 1

        for time, basal in self.basal_timeline.items():
            if start_date <= time <= end_date:
                rate = basal.get_value()
                amount_delivered = rate * basal.get_duration_hours()
                insulin_stats["total_basal"] += amount_delivered
                insulin_stats["num_basal_events"] += 1

        insulin_stats["total_insulin"] = insulin_stats["total_bolus"] + insulin_stats["total_basal"]

        if insulin_stats["num_bolus_events"] + insulin_stats["num_basal_events"] == 0:
            insulin_stats["total_bolus"] = np.nan
            insulin_stats["total_basal"] = np.nan

        return insulin_stats

    def get_carb_stats(self, start_date, end_date):
        """
        Get the sum of carbs with two datetimes, inclusive.
        Args:
            start_date (dt.DateTime): start date
            end_date (dt.DateTime): end date

        Returns:
            (float, int): total carbs and number of carb events
        """
        carb_stats = {
            "total_carbs": 0.0,
            "num_carb_events": 0
        }

        for time, food in self.food_timeline.items():
            if start_date <= time <= end_date:
                carb_stats["total_carbs"] += food.get_value()
                carb_stats["num_carb_events"] += 1

        return carb_stats

    def get_cgm_stats(self, start_date, end_date):
        """
        Compute cgm stats with dates

        Args:
            start_date (dt.DateTime): start date
            end_date (dt.DateTime): end date

        Returns:
            (float, float): geo mean and std
        """

        stats = {
            "cgm_geo_mean": np.nan,
            "cgm_geo_std": np.nan,
            "cgm_mean": np.nan,
            "cgm_std": np.nan,
            "cgm_percent_available": np.nan
        }
        cgm_values = []
        for time, cgm_event in self.cgm_timeline.items():
            if start_date <= time <= end_date:
                cgm_value = cgm_event.get_value()
                cgm_values.append(cgm_value)

        if len(cgm_values) > 1:

            seconds_per_cgm_value = (end_date - start_date).total_seconds() / 300

            stats.update({
                "cgm_geo_mean": gmean(cgm_values),
                "cgm_geo_std": gstd(cgm_values),
                "cgm_mean": np.nanmean(cgm_values),
                "cgm_std": np.nanstd(cgm_values),
                "cgm_percent_available": len(cgm_values) / seconds_per_cgm_value,
                "cgm_diff": cgm_values[0] - cgm_values[-1],
            })

        return stats

    def detect_circadian_hr(self, start_time=dt.datetime.min, end_time=dt.datetime.max, win_radius=3):
        """
        Count carb intake per hour and use the minimum as a likely cutoff for daily circadian
        boundary. Useful for window analysis.

        Args:
            start_time: datetime
                The start date of projects to use for detection

            end_time: datetime
                The end date of projects to use for detection

        Returns:
            int: hour of least user_obj activity
        """

        # Food Entries
        hour_ctr = defaultdict(int)
        if len(self.food_timeline) > 0:
            for dt, carb in self.food_timeline.items():
                if start_time <= dt <= end_time:

                    for radius in range(-win_radius, win_radius + 1):
                        associated_hour = (dt.hour + radius) % 24
                        hour_ctr[associated_hour] += 1

        # Wizard Entries
        if len(self.wizard_timeline) > 0:
            for dt, wiz in self.wizard_timeline.items():
                if start_time <= dt <= end_time:

                    for radius in range(-win_radius, win_radius + 1):
                        associated_hour = (dt.hour + radius) % 24
                        hour_ctr[associated_hour] += 1

        if len(hour_ctr) == 0:
            min_hr, min_count = 0, 0
        else:
            min_hr, min_count = min(hour_ctr.items(), key=itemgetter(1))

        return min_hr

    def detect_circadian_bg_velocity(self, start_time=dt.datetime.min, end_time=dt.datetime.max, win_radius=3):
        """
        Hour at which sum of positive bg velocity is the minimum, ie least likely time carb or insulin effects.
        """
        hour_ctr = defaultdict(int)
        if len(self.cgm_timeline) > 0:
            prev_val = None
            prev_dt = None
            for dt, cbg in self.cgm_timeline.items():
                if start_time <= dt <= end_time:

                    if prev_val:
                        velocity = (cbg.value - prev_val) / (dt - prev_dt).total_seconds()
                        if velocity > 0:
                            for radius in range(-win_radius, win_radius + 1):
                                associated_hour = (dt.hour + radius) % 24
                                hour_ctr[associated_hour] += velocity

                        prev_val = cbg.value
                        prev_dt = dt
                    else:
                        prev_val = cbg.value
                        prev_dt = dt

        if len(hour_ctr) == 0:
            min_hr, min_count = 0, 0
        else:
            min_hr, min_count = min(hour_ctr.items(), key=itemgetter(1))

        return min_hr

    def compute_window_stats(self, start_date, end_date, use_circadian=True, window_size_hours=24, hop_size_hours=24, plot_hop_raw=False):
        """
        Compute stats for a user_obj in a specified window.

        Args:
            start_date (dt.DateTime): start date
            end_date (dt.DateTime): end date
            use_circadian (bool): Use circadian hour instead of timestamp midnight for day boundary

        Returns:
            pd.DataFrame: rows are days, columns are stats
        """
        circadian_hour = 0
        if use_circadian:
            circadian_hour = self.detect_circadian_hr(win_radius=9)
            circadian_hour_bg = self.detect_circadian_bg_velocity(win_radius=5)

            logger.debug("Circadian Hour Activity {}. Circadian Hour BG Velocity {}".format(circadian_hour, circadian_hour_bg))
            circadian_hour = circadian_hour_bg

        dates = []

        start_datetime_withoffset = dt.datetime(year=start_date.year, month=start_date.month, day=start_date.day,
                                                hour=circadian_hour)

        all_window_stats = []
        segment = 0

        window_start_datetime = start_datetime_withoffset
        window_end_datetime = window_start_datetime + dt.timedelta(hours=window_size_hours)

        while True:

            dates.append(window_start_datetime.date())

            window_stats = {
                "segment_idx": segment,
                "start_date": window_start_datetime,
                "end_date": window_end_datetime
            }

            insulin_stats = self.get_insulin_stats(window_start_datetime, window_end_datetime)
            carb_stats = self.get_carb_stats(window_start_datetime, window_end_datetime)
            cgm_stats = self.get_cgm_stats(window_start_datetime, window_end_datetime)

            window_stats.update(insulin_stats)
            window_stats.update(carb_stats)
            window_stats.update(cgm_stats)

            all_window_stats.append(window_stats)

            if plot_hop_raw:
                plot_raw_data(self, window_start_datetime, window_end_datetime)

            # Update
            segment += 1
            window_start_datetime += dt.timedelta(hours=hop_size_hours)
            window_end_datetime = window_start_datetime + dt.timedelta(hours=window_size_hours)

            if window_end_datetime > end_date:
                break

        return all_window_stats

    def get_nearest_bg(self, lookup_time):
        pass

    def get_count_basal_event_types(self):
        counts = defaultdict(int)
        for _, event in self.basal_timeline.items():
            counts[event.delivery_type] += 1
        return counts

    def describe_timeline(self, timeline):

        num_events = len(timeline)
        event_dates = list(timeline.keys())
        num_days = 0
        num_events_per_day = 0
        max_events_any_day = 0
        min_events_any_day = 0
        if num_events > 1:

            dates = defaultdict(int)
            for event_datetime, event in timeline.items():
                dates[event_datetime.date()] += 1

            num_days = (event_dates[0] - event_dates[-1]).total_seconds() / 60 / 60 / 24
            num_events_per_day = num_events / num_days
            max_events_any_day = max(dates.values())
            min_events_any_day = min(dates.values())

        stats = {
            "num_events": num_events,
            "num_days": num_days,
            "num_events_per_day": num_events_per_day,
            "max_events_any_day": max_events_any_day,
            "min_events_any_day": min_events_any_day
        }
        return stats

    def describe(self):

        timelines_to_describe = [
            ("basal", self.basal_timeline),
            ("bolus", self.bolus_timeline),
            ("food", self.food_timeline),
            ("glucose", self.cgm_timeline),
            ("notes", self.note_timeline)
        ]

        stats_dict = dict()

        for name, timeline in timelines_to_describe:

            timeline_stats = self.describe_timeline(timeline)

            stats_dict[name] = timeline_stats

        stats_dict["glucose"]["cbg_percent_coverage"] = stats_dict["glucose"]["num_events_per_day"] / 288.0 * 100

        return stats_dict

    def is_keyword_in_data(self, keyword):

        events = []
        for event in self.device_data_json:
            field = get_recursively(event, keyword)
            if field:
                events.append((event, field))

        return events





import matplotlib.pyplot as plt

import seaborn as sns
sns.set_style("darkgrid")


def plot_raw_data(user, start_date, end_date):
    """
    Plot Tidepool data in its original form, ie 5-min cgm time resolution

    Args:
        user: Tidepool_User
        start_date (dt.DateTime): start date to plot
        end_date (dt.DateTime): end date to plot
    """
    fig, ax = plt.subplots(3, 1, figsize=(12, 15), sharex=True)

    event_times = []
    cgm_values = []

    if len(user.cgm_timeline) > 0:

        for dt, cgm_event in user.cgm_timeline.items():
            if start_date <= dt <= end_date:
                event_times.append(dt)
                cgm_values.append(cgm_event.get_value())

        ax[0].scatter(event_times, cgm_values)
        ax[0].set_title("CGM")
        ax[0].set_ylabel("mg/dL")

    if len(user.note_timeline) > 0:
        note_times = []
        note_values = []
        for dt, note_event in user.note_timeline.items():
            if start_date <= dt <= end_date:
                note_times.append(dt)
                note_values.append(max(cgm_values))
        ax[0].plot(note_times, note_values, label="Notes", linestyle="None", marker="s", color="y")

    ax[0].legend()

    if len(user.bolus_timeline) > 0:

        event_times = []
        bolus_values = []

        for dt, bolus_event in user.bolus_timeline.items():
            if start_date <= dt <= end_date:
                event_times.append(dt)
                bolus_values.append(bolus_event.get_value())

        if len(bolus_values) > 0:
            ax[1].set_title("Bolus")
            ax[1].stem(event_times, bolus_values)
            ax[1].set_ylabel("Units")

    if len(user.food_timeline) > 0:

        event_times = []
        carb_values = []

        for dt, carb_event in user.food_timeline.items():
            if start_date <= dt <= end_date:
                event_times.append(dt)
                carb_values.append(carb_event.get_value())

        if len(carb_values) > 0:
            ax[2].stem(event_times, carb_values)
            ax[2].set_title("Carbs")
            ax[2].set_ylabel("Grams")

    plt.title(start_date.strftime("%Y-%m-%d"))
    plt.show()


def plot_daily_stats(daily_df):
    """
    Make a plot of daily info.

    Args:
        aggregated_df pd.DataFrame: rows are days and columns are stats
    """

    stats_to_plot = [
        "cgm_geo_mean",
        "cgm_geo_std",
        "total_insulin",
        "total_carbs",
        "carb_insulin_ratio",
        "cgm_mean_insulin_ratio",
        "cgm_mean_insulin_factor"
    ]

    fig, ax = plt.subplots(len(stats_to_plot), 1, figsize=(8, 10))

    for i, stat in enumerate(stats_to_plot):
        ax[i].bar(daily_df["date"], daily_df[stat])
        ax[i].set_title(stat)

    plt.show()


def plot_daily_stats_distribution(all_daily_df):
    stats_to_plot = [
        "cgm_geo_mean",
        "cgm_geo_std",
        # "total_insulin",
        # "total_carbs",
        # "carb_insulin_ratio",
        "cgm_geo_mean_norm",
        # "total_insulin_norm",
        # "total_carbs_norm",
        # "insulin_needs"
    ]

    fig, axs = plt.subplots(nrows=len(stats_to_plot), figsize=(10, 15))

    for i, stat in enumerate(stats_to_plot):
        sns.lineplot(x="date_idx", y=stat, data=all_daily_df, hue="user_id", ax=axs[i])

    plt.show()


__author__ = "Cameron Summers"

import datetime as dt
import re

import logging
import numpy as np

import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt

import os
THIS_DIR = os.path.dirname(__file__)

PHI_DATA_DIR = os.path.join("./PHI/")


def get_logger(name):

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    return logger

logger = get_logger(__name__)

class TidepoolAPIDateParsingException(Exception):
    pass


DATESTAMP_FORMAT = "%Y-%m-%d"
API_DATA_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def parse_tidepool_api_date_str(date_str):
    """
    Parse date strings in formats common to Tidepool API

    Args:
        date_str (str): date string

    Returns:
        dt.DateTime
    """
    common_timestamp_formats = [
        API_DATA_TIMESTAMP_FORMAT,
        "%Y-%m-%dT%H:%M:%SZ",
        DATESTAMP_FORMAT
    ]

    datetime_obj = None

    # Some devices have 7 zeros instead of six, which datetime can't handle.
    if len(date_str) == len('2021-03-24T14:05:29.0000000Z'):
        date_str = re.sub("\d{7}Z", "000000Z", date_str)
    elif len(date_str) == len('2021-03-24T14:05:29.00000000Z'):
        date_str = re.sub("\d{8}Z", "000000Z", date_str)
    elif len(date_str) == len('2021-03-24T14:05:29.000000000Z'):
        date_str = re.sub("\d{9}Z", "000000Z", date_str)

    try:
        datetime_obj = dt.datetime.fromisoformat(date_str)
    except ValueError:
        for format in common_timestamp_formats:

            try:
                datetime_obj = dt.datetime.strptime(date_str, format)
            except:
                pass

    if datetime_obj is None:
        raise TidepoolAPIDateParsingException("String '{}' could not be parsed.".format(date_str))

    # Notes have
    if datetime_obj.tzinfo is not None:
        offset = datetime_obj.utcoffset()
        datetime_obj = datetime_obj + offset
        datetime_obj = datetime_obj.replace(tzinfo=None)

    return datetime_obj


def get_user_group_data_dir(user_group_name):

    return os.path.join(PHI_DATA_DIR, "tidepool_user_groups", user_group_name)


def create_user_dir(user_id, start_date, end_date, user_group_name=""):
    """
    Create
    Args:
        start_date dt.DateTime: start date of data for user_obj
        end_date dt.DateTime: end date of data for user_obj
        user_id (str): user_obj id for user_obj

    Returns:
        str: dir_path for saving data
    """
    if not os.path.isdir(PHI_DATA_DIR):
        raise Exception("You are not saving to PHI folder. Check your path.")

    user_group_dir = get_user_group_data_dir(user_group_name)

    user_dir_name = "{}_{}_{}".format(user_id, start_date.strftime(DATESTAMP_FORMAT), end_date.strftime(DATESTAMP_FORMAT))
    user_dir_path = os.path.join(user_group_dir, user_dir_name)

    if not os.path.isdir(user_dir_path):
        os.makedirs(user_dir_path)

    return user_dir_path


def get_recursively(search_dict, keyword):
    """
    Takes a dict with nested lists and dicts,
    and searches all dicts for a key of the field
    provided or field in .

    ref: https://stackoverflow.com/questions/14962485/finding-a-key-recursively-in-a-dictionary
    """
    fields_found = []

    for key, value in search_dict.items():

        if key == keyword:
            fields_found.append(value)

        if isinstance(search_dict[key], str) and keyword in search_dict[key]:
            fields_found.append(value)

        elif isinstance(value, dict):
            results = get_recursively(value, keyword)
            for result in results:
                fields_found.append(result)

        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    more_results = get_recursively(item, keyword)
                    for another_result in more_results:
                        fields_found.append(another_result)


    return fields_found


def logger_describe_distribution(description, distribution):

    distribution = np.array(distribution)
    nan_mask = np.isnan(distribution)
    num_before = len(distribution)
    distribution = distribution[~nan_mask]
    num_after = len(distribution)

    logger.info("{} Before NanMask. {} After.".format(num_before, num_after))

    logger.info("{}: mean={:.2f}. min={:.2f}. q1={:.2f}. median={:.2f}. q3={:.2f}. max={:.2f}".format(
        description,
        np.nanmean(distribution),
        np.nanmin(distribution),
        np.nanpercentile(distribution, 25),
        np.nanmedian(distribution),
        np.nanpercentile(distribution, 75),
        np.nanmax(distribution),
    ))


def get_years_since_date(reference_date):

    years_since = (dt.datetime.today() - reference_date).total_seconds() / 60 / 60 / 24 / 365
    return years_since


def create_synthetic_dataset(num_days):

    data = []
    cir = 10
    target = 100
    br_total = 10
    np.random.seed(1234)
    for i in range(num_days):
        carb_trend = 1.0  #+ 0.2*np.cos(2*np.pi * i / num_days)
        total_carbs_true = np.random.normal(150, 30) * carb_trend

        # carb_estimation_std = total_carbs_true * 0.01
        carb_estimation_std = 20
        total_carbs = np.random.normal(total_carbs_true, carb_estimation_std)
        carb_diff = (total_carbs_true - total_carbs)

        insulin_sensitivity = 1.0 #+ 0.2*np.sin(2*np.pi * i / num_days)
        total_insulin = br_total + total_carbs_true / cir * insulin_sensitivity

        tir = 1.0 - abs(carb_diff) / 100
        day = {
            "date": dt.datetime(2021, 1, 1, 0, 0, 0) + dt.timedelta(days=i),
            "total_carbs_true": total_carbs_true,
            "total_carbs": total_carbs,
            "total_insulin": total_insulin,
            "cgm_geo_mean": target * insulin_sensitivity,
            "cgm_mean": target * insulin_sensitivity,
            "cgm_percent_below_54": 0,
            "cgm_percent_tir": tir,
            "carb_diff": carb_diff
        }
        data.append(day)

    df = pd.DataFrame(data)

    window_size_days = 60
    indexer = pd.api.indexers.FixedForwardWindowIndexer(window_size=window_size_days)
    # df_agg = df.rolling(window=indexer, min_periods=1).mean()
    # df_agg = df_agg.iloc[:-window_size_days]
    # df_agg["date"] = df["date"].iloc[:-window_size_days]
    # df = df_agg

    g = sns.pairplot(df[["total_carbs_true", "total_carbs", "total_insulin", "cgm_percent_tir"]])

    df.to_csv("synthetic_aggregated_dataset.csv")

    plt.title('Fake Dataset')
    plt.show()

    return df


if __name__ == "__main__":
    create_synthetic_dataset(60)